/*  SciGraphica - Scientific graphics and data manipulation
 *  Copyright (C) 2000 Adrian E. Feiguin <feiguin@ifir.edu.ar>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <gtkextra/gtkextra.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <string.h>
#include <Python.h>
#include <arrayobject.h>
#include "main.h"
#include "python/python_int.h"
#include "sg_worksheet_tools.h"


#define DEFAULT_PRECISION 3

static void	sg_worksheet_unmap			(GtkWidget *window, 
							 GdkEvent *event);

static void	sg_worksheet_destroy			(GtkWidget *window); 
static gint 	sg_worksheet_delete			(GtkWidget *window); 
static gboolean activate_cell				(GtkSheet *sheet, 
							 gint row, gint col,
							 gpointer data);
static gboolean deactivate_cell                         (GtkSheet *sheet,
                                                         gint row, gint col,
                                                         gpointer data);

static void clear_cell                         		(GtkSheet *sheet,
                                                         gint row, gint col,
                                                         gpointer data);
static gboolean focus_in				(GtkWidget *widget,
							 GdkEventFocus *event);

/* Global string */
gchar *cell_save = NULL;
extern GdkPixmap *icon_pixmap;
extern GdkBitmap *icon_mask;

SGworksheet *
sg_worksheet_new(gchar *name, gint nrows, gint ncols)
{
  SGworksheet *worksheet;
  GtkSheet *sheet;
  gint c;

  worksheet = g_new(SGworksheet, 1);

  worksheet->sheet = gtk_sheet_new(nrows, ncols, name);
  gtk_sheet_set_row_titles_width(GTK_SHEET(worksheet->sheet), 60);
  GTK_SHEET_UNSET_FLAGS(GTK_SHEET(worksheet->sheet), GTK_SHEET_JUSTIFY_ENTRY);

  worksheet->last_column = ncols-1;

  sheet = GTK_SHEET(worksheet->sheet);
  GTK_SHEET_SET_FLAGS(sheet, GTK_SHEET_ROW_FROZEN);

  worksheet->column = g_new(SGcolumn, 5);

  for(c = 0; c < ncols; c++){
     gchar label[4];
     gint max = 'Z'-'A'+1;
     gint n;

     n = 0;
     if(c >= max*max){
        label[0] = c/(max*max)+'A'-1;
        label[1] = '\0';
        n = 1;
        c -= (c/(max*max))*max*max;
     }
     if(c >= max){
        label[n] = c/max+'A'-1;
        label[n+1] = '\0';
        n++ ;
        c -= (c/max)*max;
     }
     if(c < max){
        label[n] = c+'A';
        label[n+1] = '\0';
     }

    gtk_sheet_column_button_add_label(sheet, c, label);
    gtk_sheet_set_column_title(sheet, c, label);
    sg_worksheet_column_set_format(worksheet, c, SG_COLUMN_NUMBER, 
                                   DEFAULT_PRECISION);
    worksheet->column[c].exp = NULL;
  }

  worksheet->name = g_strdup(name);
 
  worksheet->window = NULL;
  worksheet->sw = NULL;
  worksheet->is_mapped = FALSE;

  worksheet->x = 0;
  worksheet->y = 0;
  worksheet->width = 400;
  worksheet->height = 350;

  worksheet->begin = -1;
  worksheet->end = -1;

  gtk_signal_connect(GTK_OBJECT(worksheet->sheet), "activate",
                     GTK_SIGNAL_FUNC(activate_cell), worksheet);

  gtk_signal_connect(GTK_OBJECT(worksheet->sheet), "deactivate",
                     GTK_SIGNAL_FUNC(deactivate_cell), worksheet);

  gtk_signal_connect(GTK_OBJECT(worksheet->sheet), "clear_cell",
                     GTK_SIGNAL_FUNC(clear_cell), worksheet);

  return worksheet;
}   

void
sg_worksheet_column_set_format(SGworksheet *worksheet,
                               gint column, 
                               SGcolumnformat format,
                               gint precision)
{
  GtkSheet *sheet;

  sheet = GTK_SHEET(worksheet->sheet);

  worksheet->column[column].format = format;
  worksheet->column[column].precision = precision;

  switch(worksheet->column[column].format){
    case SG_COLUMN_NUMBER:
       gtk_sheet_column_set_justification(sheet, column, GTK_JUSTIFY_RIGHT);
       break;
    case SG_COLUMN_DATE:
    case SG_COLUMN_TIME:
       gtk_sheet_column_set_justification(sheet, column, GTK_JUSTIFY_CENTER);
       break;
    case SG_COLUMN_TEXT:
    default:
       gtk_sheet_column_set_justification(sheet, column, GTK_JUSTIFY_LEFT);
  }

   
}

void 
sg_worksheet_cell_clear(SGworksheet *sheet, gint row, gint col)
{ gtk_sheet_cell_clear(GTK_SHEET(sheet->sheet), row, col);
}

static void 
clear_cell(GtkSheet *sheet, gint row, gint col, gpointer data)
{
  SGhiddencell *hidden;

  hidden = (SGhiddencell *)gtk_sheet_get_link(sheet, row, col);
  if(hidden && hidden->formula) g_free(hidden->formula);

}

static gboolean
activate_cell(GtkSheet *sheet, gint row, gint col, gpointer data)
{
  SGworksheet *worksheet;
  GtkEntry *entry;
  gchar *text = NULL;
  gchar label[100];
  SGhiddencell *link = NULL;

  entry = GTK_ENTRY(sheet->sheet_entry);
  worksheet = (SGworksheet *)data;

  cell_save = g_strdup(gtk_entry_get_text(entry));
  if(worksheet->column[col].format == SG_COLUMN_TEXT) return;

  link=(SGhiddencell *)gtk_sheet_get_link(GTK_SHEET(worksheet->sheet),row,col);

  if (link)
   {  
      text = link->formula;
      if(text) gtk_entry_set_text(entry, text);
   }
  else
   { 
      text = NULL;
      gtk_entry_set_text(entry,"");
   }

  if(text && strlen(text)){
    gtk_entry_select_region(entry, 0, strlen(text));
    gtk_entry_set_position(entry, strlen(text));
  }

  sprintf(label, "%s:%d", sheet->column[col].name, row);
  gtk_label_set(GTK_LABEL(worksheet->label), label);

  return TRUE;
}


static gboolean
deactivate_cell(GtkSheet *sheet, gint row, gint col, gpointer data)
{
  SGworksheet *worksheet;
  gchar *text;


  worksheet = (SGworksheet *)data;

  text = gtk_entry_get_text(GTK_ENTRY(sheet->sheet_entry));

  sg_worksheet_cell_set(worksheet, row, col, text, TRUE, TRUE);

  return TRUE;
}

void
sg_worksheet_cell_set(SGworksheet *worksheet, gint row, gint col,
                      gchar *text, gboolean formula, gboolean eval)
{
  gchar *save;
  SGhiddencell *link = NULL;

  if(worksheet->column[col].format == SG_COLUMN_TEXT){
     return;
  }

  link = (SGhiddencell *)gtk_sheet_get_link(GTK_SHEET(worksheet->sheet), row, col);

  if (text && strlen(text))
   { 
     if (!link && formula)
       { 
         link = (SGhiddencell *)g_new0(SGhiddencell, 1);
         link->formula = NULL;
         link->updated = FALSE;
         gtk_sheet_link_cell(GTK_SHEET(worksheet->sheet), row, col, link);
       }

     save = g_strdup(text);

     if (!formula)
         gtk_sheet_set_cell_text(GTK_SHEET(worksheet->sheet),
                                 row, col, save);
     else
     {  
       link->updated=FALSE;
       if (eval)
       { /* Removed this test as an experiment to improve the GUI.
            CDS 27/11/2000
          if (link && link->formula && !strcmp(link->formula, text))
          {
            gtk_sheet_set_cell_text(GTK_SHEET(worksheet->sheet),
                                    row, col, (cell_save?cell_save:save));
            link->updated=TRUE;
          }
          else*/
          {
              if (python_sheet(worksheet, row, col, save, GTK_ORIENTATION_VERTICAL))
                  link->updated=TRUE;
          }
        }
       else
          gtk_sheet_set_cell_text(GTK_SHEET(worksheet->sheet),
                                  row, col, (save?save:text));

       if (link->formula)
          g_free(link->formula);
       link->formula = save;
     }
   }
  else /* We were passed an empty cell */
   { 
     if (link)
      {
        if (link->formula) g_free(link->formula);
        gtk_sheet_remove_link(GTK_SHEET(worksheet->sheet), row, col);
        g_free(link);
        gtk_sheet_cell_clear(GTK_SHEET(worksheet->sheet), row, col);
      }
     else
        gtk_sheet_set_cell_text(GTK_SHEET(worksheet->sheet),
                                row, col, g_strdup(cell_save));
   }

  if(cell_save) g_free(cell_save);
  cell_save = NULL;

}


gchar *
sg_worksheet_cell_get_text(SGworksheet *worksheet, gint row, gint col)
{
  gchar *text;

  text = gtk_sheet_cell_get_text(GTK_SHEET(worksheet->sheet), row, col);

  return text;
}

gchar *
sg_worksheet_cell_get_formula(SGworksheet *worksheet, gint row, gint col)
{
  gchar *text=NULL;
  SGhiddencell *link;

  link = (SGhiddencell *)gtk_sheet_get_link(GTK_SHEET(worksheet->sheet),
                                            row, col);

  if(link) text = link->formula;

  return text;
}


  
void
sg_worksheet_open(SGworksheet *worksheet)
{
  gchar msg[80];
  GtkWidget *vbox;
  GtkWidget *frame;
  GtkWidget *toolbar;
#ifndef WITH_GNOME
  GtkWidget *hb;
#endif

  active_worksheet=worksheet;

#ifdef WITH_GNOME
    snprintf(msg,80,"Active sheet is %s", active_worksheet->name);
    gnome_app_flash(GNOME_APP(main_window), msg);
#else
/* Do something equivalent for gtk!*/
#endif

  if(worksheet->is_mapped) {
       gdk_window_raise(worksheet->window->window);
       return;
  }

  vbox = gtk_vbox_new(FALSE, 5);

#ifdef WITH_GNOME
  worksheet->window = gnome_app_new (PACKAGE, worksheet->name);
  gnome_app_set_contents(GNOME_APP(worksheet->window),vbox);

  toolbar = sg_worksheet_build_toolbar(worksheet);
#else
  worksheet->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_add(GTK_CONTAINER(worksheet->window),vbox);

  hb = gtk_handle_box_new();
  gtk_handle_box_set_handle_position(GTK_HANDLE_BOX(hb), GTK_POS_LEFT);
  gtk_box_pack_start(GTK_BOX(vbox), hb, FALSE, FALSE, 0);
  gtk_widget_show(hb);

  toolbar = sg_worksheet_build_toolbar(worksheet);
  gtk_container_add(GTK_CONTAINER(hb), toolbar);
  gtk_widget_show_all(toolbar);
#endif

  frame = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
  worksheet->label = gtk_label_new(" ");
  gtk_misc_set_alignment(GTK_MISC(worksheet->label), 0., .5);
  gtk_container_add(GTK_CONTAINER(frame), worksheet->label);
 
  worksheet->sw = gtk_scrolled_window_new(NULL, NULL);
 
  gtk_window_set_title(GTK_WINDOW(worksheet->window), worksheet->name);
  gtk_window_set_policy(GTK_WINDOW(worksheet->window), TRUE, TRUE, FALSE);
 
  if(worksheet->x > 0 && worksheet->y > 0){
           gtk_widget_set_uposition(worksheet->window, 
                                    worksheet->x, worksheet->y);
           gtk_widget_set_usize(worksheet->window,
                                worksheet->width,
                                worksheet->height);
  }else{
           gtk_widget_set_usize(worksheet->window,
                                400,
                                350);
  }
 
  gtk_signal_connect(GTK_OBJECT(worksheet->window),"unmap_event",
                    (GtkSignalFunc) sg_worksheet_unmap, NULL);
 
  gtk_signal_connect(GTK_OBJECT(worksheet->window),"destroy",
                     (GtkSignalFunc) sg_worksheet_destroy, NULL);
 
  gtk_signal_connect(GTK_OBJECT(worksheet->window),"delete_event",
                     (GtkSignalFunc) sg_worksheet_delete, NULL);
 
  gtk_signal_connect(GTK_OBJECT(worksheet->window),"focus_in_event",
                     (GtkSignalFunc) focus_in, NULL);
 
  gtk_container_add(GTK_CONTAINER(worksheet->sw), worksheet->sheet);
  gtk_box_pack_start(GTK_BOX(vbox), worksheet->sw, TRUE, TRUE, 0);
  gtk_box_pack_end(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  
  gtk_widget_ensure_style(worksheet->sheet);
  gtk_widget_realize(worksheet->window);
  gdk_window_set_icon(worksheet->window->window, NULL, icon_pixmap, icon_mask);             
  gtk_widget_show(worksheet->window);
  gtk_widget_show(worksheet->sw);
  gtk_widget_show(worksheet->sheet);
  gtk_widget_show(worksheet->label);
  gtk_widget_show(frame);
  gtk_widget_show(vbox);
  gtk_widget_ref(worksheet->sheet);
  worksheet->is_mapped = TRUE;
}

void
sg_worksheet_hide(SGworksheet *worksheet)
{
  GList *list;
  gchar msg[80];

/* Get another open sheet in list */
  list=worksheets;
  while (list)
  { 
    if (((SGworksheet*)list->data)->is_mapped &&
       list->data != (gpointer)worksheet) break;
    list=list->next;
  }
  if (list)
  { 
    active_worksheet=(SGworksheet*)(list->data);
#ifdef WITH_GNOME
    snprintf(msg,80,"Active sheet is %s", active_worksheet->name);
    gnome_app_flash(GNOME_APP(main_window),msg);
#else
/* Do something equivalent for gtk!*/
#endif
  }
  else
  { 
    active_worksheet = NULL;
#ifdef WITH_GNOME
    gnome_app_flash(GNOME_APP(main_window), "No active sheet");
#else
/* Do something equivalent for gtk!*/
#endif
  }

  gtk_container_remove(GTK_CONTAINER(worksheet->sw), worksheet->sheet);
  if(GTK_IS_WIDGET(worksheet->sw)) gtk_widget_destroy(worksheet->sw);
  if(GTK_OBJECT(worksheet->sheet)->ref_count > 1) 
                     gtk_widget_unref(worksheet->sheet);

  worksheet->is_mapped = FALSE;
  gtk_widget_destroy(worksheet->window);
  worksheet->sw = NULL;
  worksheet->window = NULL;
  worksheet->label = NULL;
}

gint
sg_worksheet_rename(SGworksheet *worksheet, gchar *name)
{

  if(worksheet->name){
     g_free(worksheet->name);
     worksheet->name = NULL;
  }
 
  worksheet->name = g_strdup(name);

  if(worksheet->window) 
     gtk_window_set_title(GTK_WINDOW(worksheet->window),name);

  if(worksheet->icon->entry_label){
     g_free(worksheet->icon->entry_label);
     worksheet->icon->entry_label = NULL;
  }
 
  worksheet->icon->entry_label = g_strdup(name);

  if(worksheet->icon->entry_label){
     g_free(worksheet->icon->label);
     worksheet->icon->label = NULL;
  }
 
  worksheet->icon->label = g_strdup(name);

  gtk_entry_set_text(GTK_ENTRY(worksheet->icon->entry), name);

  return TRUE;
}


static void
sg_worksheet_unmap(GtkWidget *window, GdkEvent *event)
{
  GList *w;
  SGworksheet *worksheet = NULL;
  GtkWidget *data;

  w = worksheets;
  while(w){
    worksheet = (SGworksheet *)w->data;
    data = worksheet->window;
    if(data && data == window) break;
    w = w->next;
  }

  gdk_window_get_origin(worksheet->window->window, 
                          &worksheet->x,
                          &worksheet->y);
  gdk_window_get_size(worksheet->window->window, 
                          &worksheet->width,
                          &worksheet->height);

  sg_worksheet_hide(worksheet);

}

static void
sg_worksheet_destroy(GtkWidget *window)
{
  GList *w;
  SGworksheet *worksheet = NULL;
  GtkWidget *data;

  w = worksheets;
  while(w){
    worksheet = (SGworksheet *)w->data;
    data = worksheet->window;
    if(data && data == window) break;
    w = w->next;
  }

  if(!worksheet || !worksheet->window) return;
  if(!GTK_WIDGET_MAPPED(window) && worksheet->is_mapped){
            sg_worksheet_hide(worksheet);
  }
}

static gint
sg_worksheet_delete(GtkWidget *window)
{
  GList *w;
  SGworksheet *worksheet = NULL;
  GtkWidget *data;

  w = worksheets;
  while(w){
    worksheet = (SGworksheet *)w->data;
    data = worksheet->window;
    if(data && data == window) break;
    w = w->next;
  }

  gdk_window_get_origin(worksheet->window->window, 
                          &worksheet->x,
                          &worksheet->y);
  gdk_window_get_size(worksheet->window->window, 
                          &worksheet->width,
                          &worksheet->height);

  sg_worksheet_hide(worksheet);

  return FALSE;
}

void
sg_worksheet_set_begin(SGworksheet *worksheet, gint row)
{
  GtkSheet *sheet;
  GdkColor bg;
  GtkSheetRange range;
  gchar label[100];


  sheet = GTK_SHEET(active_worksheet->sheet);

  if(row > worksheet->end &&
     worksheet->end != -1) return;

  gtk_sheet_freeze(sheet);

  if(worksheet->begin != -1){
    range.col0 = 0;
    range.coli = sheet->maxcol;
    range.row0 = worksheet->begin;
    range.rowi = worksheet->begin;
    gdk_color_white(gdk_colormap_get_system(), &bg);
    gtk_sheet_row_button_add_label(sheet, worksheet->begin, NULL);
    gtk_sheet_range_set_background(sheet, &range, &bg);
  }

  gdk_color_parse("light blue", &bg);
  gdk_color_alloc(gdk_colormap_get_system(), &bg);
  worksheet->begin = row;
  range.col0 = 0;
  range.coli = sheet->maxcol;
  range.row0 = row;
  range.rowi = row;

  sprintf(label, "(Begin)");

  if(row > -1){
    gtk_sheet_row_button_add_label(sheet, row, label);
    gtk_sheet_range_set_background(sheet, &range, &bg);
  }
  gtk_sheet_thaw(sheet);
}

void
sg_worksheet_set_end(SGworksheet *worksheet, gint row)
{
  GtkSheet *sheet;
  GdkColor bg;
  GtkSheetRange range;
  gchar label[100];


  sheet = GTK_SHEET(active_worksheet->sheet);

  if(row != -1 && row < worksheet->begin) return;

  gtk_sheet_freeze(sheet);

  if(worksheet->end != -1){
    range.col0 = 0;
    range.coli = sheet->maxcol;
    range.row0 = worksheet->end;
    range.rowi = worksheet->end;
    gdk_color_white(gdk_colormap_get_system(), &bg);
    gtk_sheet_row_button_add_label(sheet, worksheet->end, NULL);
    gtk_sheet_range_set_background(sheet, &range, &bg);
  }

  gdk_color_parse("light blue", &bg);
  gdk_color_alloc(gdk_colormap_get_system(), &bg);
  worksheet->end = row;
  range.col0 = 0;
  range.coli = sheet->maxcol;
  range.row0 = row;
  range.rowi = row;

  sprintf(label, "(End)");

  if(row > -1){
    gtk_sheet_row_button_add_label(sheet, row, label);
    gtk_sheet_range_set_background(sheet, &range, &bg);
  }

  gtk_sheet_thaw(sheet);
}


void
sg_worksheet_reset(SGworksheet *worksheet)
{
  sg_worksheet_set_begin(worksheet, -1);
  sg_worksheet_set_end(worksheet, -1);
}

void
sg_worksheet_remove(SGworksheet *worksheet)
{
  if(worksheet->is_mapped)
     sg_worksheet_hide(worksheet);

  gtk_widget_destroy(worksheet->sheet);

  g_free(worksheet->name);
  g_free(worksheet->column);
  g_free(worksheet);
}

void
sg_worksheet_add_rows(SGworksheet *worksheet, gint nrows)
{
  GtkSheet *sheet;

  sheet = GTK_SHEET(worksheet->sheet);

  gtk_sheet_freeze(sheet);
  gtk_sheet_add_row(sheet, nrows);

  sg_worksheet_set_begin(worksheet, worksheet->begin);
  sg_worksheet_set_end(worksheet, worksheet->end);

  gtk_sheet_thaw(sheet);
}

void
sg_worksheet_add_columns(SGworksheet *worksheet, gint ncols)
{
  GtkSheet *sheet;
  gint i, c, n = 0, max;
  gchar label[4];

  sheet = GTK_SHEET(worksheet->sheet);

  gtk_sheet_freeze(sheet);

  max = 'Z'-'A'+1;

  gtk_sheet_add_column(sheet, ncols);

  worksheet->column = (SGcolumn *)g_realloc(worksheet->column, 
                                  (sheet->maxcol + 1) * sizeof(SGcolumn));

  for(i = sheet->maxcol - ncols + 1; i <= sheet->maxcol ; i++){
      worksheet->last_column++;
      c = worksheet->last_column;
      n = 0;
      if(c >= max*max){
         label[0] = c/(max*max)+'A'-1;
         label[1] = '\0';
         n = 1;
         c -= (c/(max*max))*max*max;
      }
      if(c >= max){
         label[n] = c/max+'A'-1;
         label[n+1] = '\0';
         n++ ;
         c -= (c/max)*max;
      }
      if(c < max){
         label[n] = c+'A';
         label[n+1] = '\0';
      }
      gtk_sheet_column_button_add_label(sheet, i, label);
      gtk_sheet_set_column_title(sheet, i, label);
      gtk_sheet_column_set_justification(sheet, i, GTK_JUSTIFY_RIGHT);
      sg_worksheet_column_set_format(worksheet, i, SG_COLUMN_NUMBER, 
                                     DEFAULT_PRECISION);
      worksheet->column[i].exp = NULL;
  }

  sg_worksheet_set_begin(worksheet, worksheet->begin);
  sg_worksheet_set_end(worksheet, worksheet->end);

  gtk_sheet_thaw(sheet);

}

void
sg_worksheet_insert_rows(SGworksheet *worksheet, 
                         gint row,
                         gint nrows)
{
  GtkSheet *sheet;

  sheet = GTK_SHEET(worksheet->sheet);

  gtk_sheet_insert_rows(sheet, row, nrows);

  if(worksheet->end >= row)
      sg_worksheet_set_end(worksheet, worksheet->end + nrows);

  if(worksheet->begin >= row)
      sg_worksheet_set_begin(worksheet, worksheet->begin + nrows);

}

void
sg_worksheet_insert_columns(SGworksheet *worksheet, 
                            gint col,
                            gint ncols)
{
  GtkSheet *sheet;
  gint i, c, n, max;
  gchar label[4];
  SGcolumn auxcol;

  sheet = GTK_SHEET(worksheet->sheet);

  gtk_sheet_freeze(sheet);

  gtk_sheet_insert_columns(sheet, col, ncols);

  worksheet->column = (SGcolumn *)g_realloc(worksheet->column,
                                  (sheet->maxcol + 1) * sizeof(SGcolumn));


  for(i=sheet->maxcol; i>=col+ncols; i--){
    auxcol = worksheet->column[i];
    worksheet->column[i]=worksheet->column[i-ncols];
    worksheet->column[i-ncols]=auxcol;
  }

  max = 'Z'-'A'+1;
  for(i = col; i < col+ncols; i++){
      worksheet->last_column++;
      c = worksheet->last_column;

      n = 0;
      if(c >= max*max){
         label[0] = c/(max*max)+'A'-1;
         label[1] = '\0';
         n = 1;
         c -= (c/(max*max))*max*max;
      }
      if(c >= max){
         label[n] = c/max+'A'-1;
         label[n+1] = '\0';
         n++ ;
         c -= (c/max)*max;
      }
      if(c < max){
         label[n] = c+'A';
         label[n+1] = '\0';
      }

      gtk_sheet_column_button_add_label(sheet, i, label);
      gtk_sheet_set_column_title(sheet, i, label);
      gtk_sheet_column_set_justification(sheet, i, GTK_JUSTIFY_RIGHT);
      sg_worksheet_column_set_format(worksheet, i, SG_COLUMN_NUMBER, 
                                     DEFAULT_PRECISION);
      worksheet->column[i].exp = NULL;
    }

  gtk_sheet_thaw(sheet);

}

void
sg_worksheet_delete_rows(SGworksheet *worksheet, gint row, gint nrows)
{
  GtkSheet *sheet;

  sheet = GTK_SHEET(worksheet->sheet);

  gtk_sheet_delete_rows(sheet, row,
                        nrows);

  if(worksheet->begin >= row &&
     worksheet->begin <= row+nrows)
          sg_worksheet_set_begin(worksheet, -1);

  if(worksheet->begin > row+nrows)
          sg_worksheet_set_begin(worksheet,
                                 worksheet->begin - nrows);

  if(worksheet->end >= row &&
     worksheet->end <= row+nrows)
          sg_worksheet_set_end(worksheet, -1);

  if(worksheet->end > row+nrows)
          sg_worksheet_set_end(worksheet,
                               worksheet->end - nrows);
}

void
sg_worksheet_delete_columns(SGworksheet *worksheet, gint col, gint ncols)
{
  GtkSheet *sheet;
  gint i;

  sheet = GTK_SHEET(active_worksheet->sheet);

  gtk_sheet_delete_columns(sheet, 
                           col,
                           col+ncols);


  for(i = col; i <= sheet->maxcol-ncols; i++){
    sg_worksheet_column_set_exp(worksheet, i, NULL);
    worksheet->column[i]=worksheet->column[i+ncols];
  }

}

void
sg_worksheet_column_set_exp(SGworksheet *worksheet, gint col, gchar *exp)
{
  if(worksheet->column[col].exp)
     g_free(worksheet->column[col].exp);

  worksheet->column[col].exp = NULL;

  if(exp)
     worksheet->column[col].exp = g_strdup(exp);
}

static gboolean
focus_in(GtkWidget *widget, GdkEventFocus *event)
{
  GList *list;
  SGworksheet *worksheet;

  list = worksheets;
  while(list)
   {
     worksheet = (SGworksheet *)list->data;

     if(worksheet->window == widget){
        active_worksheet = worksheet;
        break;
     }

     list = list->next;
   }

  return FALSE;

}


void
sg_worksheet_update_column_exp (SGworksheet *worksheet, 
                                gchar *exp, gint column,
                                gint from, gint to)
{ 
  PyObject *object;
  gint i;

  sg_worksheet_column_set_exp(worksheet, column, exp);

  object = (PyObject *)python_eval_expr(exp);

  if (object)
  {
      if (PyArray_Check(object))
       { 
         gtk_sheet_freeze(GTK_SHEET(worksheet->sheet));
         python_array(worksheet,0,column,(PyArrayObject *)object,
                      GTK_ORIENTATION_VERTICAL,TRUE);
         gtk_sheet_thaw(GTK_SHEET(worksheet->sheet));
       }
      else
       {
        if (PySequence_Check(object))
         { 
           gtk_sheet_freeze(GTK_SHEET(worksheet->sheet));
           python_sequence(worksheet,0,column,object,
                           GTK_ORIENTATION_VERTICAL,TRUE);
           gtk_sheet_thaw(GTK_SHEET(worksheet->sheet));
         }
        else
           for (i=from;i<=to;i++) /* inclusive */
             python_singleton(worksheet,i,column,object,TRUE);
       }
       Py_XDECREF(object);
   }

}

/* Mark all formulas in cells as not updated */
gint sg_worksheet_unupdate_exp_all(SGworksheet *worksheet)
{
  return sg_worksheet_unupdate_exp_range(worksheet, 
           0, GTK_SHEET(worksheet->sheet)->maxallocrow,
           0, GTK_SHEET(worksheet->sheet)->maxalloccol);
}                                   


/* Mark all formulas in a range as not updated */
gint sg_worksheet_unupdate_exp_range(SGworksheet *worksheet,
                                   gint row0, gint rowi,
                                   gint col0, gint coli)
{ SGcolumn *column;
  gint col,row;
  gchar *exp;
  SGhiddencell *link = NULL;


  link = (SGhiddencell *)gtk_sheet_get_link(GTK_SHEET(worksheet->sheet),
                                            row, col);

  for(row=row0;row<=rowi;row++)
      for (col=col0;col<=coli;col++)
      {   exp=sg_worksheet_cell_get_formula(worksheet, row, col);
          if (!exp) continue;
          link = (SGhiddencell *)gtk_sheet_get_link(GTK_SHEET(worksheet->sheet), row, col);
          if (link) link->updated=FALSE;
      }
  return TRUE;
}

gint sg_worksheet_update_exp_all(SGworksheet *worksheet)
{
  return sg_worksheet_update_exp_range(worksheet, 
           0, GTK_SHEET(worksheet->sheet)->maxallocrow,
           0, GTK_SHEET(worksheet->sheet)->maxalloccol);

}                                   
gint sg_worksheet_update_exp_range(SGworksheet *worksheet, 
                                   gint row0, gint rowi,
                                   gint col0, gint coli)
{ SGcolumn *column;
  gint col,row,uneval1=-1,uneval2=0;
  gchar *exp, *exp2;
  SGhiddencell *link = NULL;
  gint srow,scol,num_selected=0;
  
  gtk_sheet_freeze(GTK_SHEET(worksheet->sheet));

  gtk_sheet_get_active_cell(GTK_SHEET(worksheet->sheet),&srow,&scol);  
  num_selected=
  ((GTK_SHEET(worksheet->sheet)->range.rowi-GTK_SHEET(worksheet->sheet)->range.row0)+1)*
  ((GTK_SHEET(worksheet->sheet)->range.coli-GTK_SHEET(worksheet->sheet)->range.col0)+1);


  /* First update all the column formulas */
  for (col=col0;col<=coli;col++)
    if (worksheet->column[col].exp)
    {   exp=strdup(worksheet->column[col].exp);
        sg_worksheet_update_column_exp(worksheet,exp,col,0,
          gtk_sheet_get_rows_count(GTK_SHEET(worksheet->sheet)) - 1);
        g_free(exp);
    }

  sg_report_python_error=FALSE;
  while (uneval1!=uneval2)
  {   /* Report problems from the second pass */
      if (uneval1>=0) sg_report_python_error=TRUE;
      uneval1=uneval2;
      for(row=row0,uneval2=0;row<=rowi;row++)
          for (col=col0;col<=coli;col++)
          {   exp=sg_worksheet_cell_get_formula(worksheet, row, col);
              if (!exp) continue;
              link = (SGhiddencell *)gtk_sheet_get_link(GTK_SHEET(worksheet->sheet), row, col);
              if (link->updated) continue;
              exp2=g_strdup(exp);
              sg_worksheet_cell_set(worksheet, row, col,exp,TRUE,TRUE);
              if (row==srow && col==scol && num_selected==1)
                 gtk_entry_set_text(GTK_ENTRY(GTK_SHEET(worksheet->sheet)->sheet_entry),
                                   exp2);
              g_free(exp2);                
              if (!link->updated) uneval2++;
          }
       if (uneval2==0) break;
  }
  gtk_sheet_thaw(GTK_SHEET(worksheet->sheet));
  return uneval2;
}

